#ifndef _SP_H
#define _SP_H

#ifndef _NEW_
	#include <new>
#endif

#ifndef _CONVERSION_H
	#include "conversion.h"
#endif

template <class T, class R, template <class> class S, template <class> class GC, template <class, template <class> class> class W> class WH;

template <class R, template <class> class S, template <class> class GC, template <class, template <class> class> class W>
class AbstractSP : public R, public S<AbstractSP>
{
	friend class S<AbstractSP>;
	friend class W<AbstractSP, GC>;
public:
	typedef S<AbstractSP> SPAllocator;
	typedef W<AbstractSP, GC> ObjectAllocator;

	AbstractSP(void* object, size_t size) : d_object(object), d_size(size), d_active(true){}
	virtual ~AbstractSP()
	{
		d_object = 0;
		d_size = 0;
	}

	const void* const getObject() const
	{
		return d_object;
	}
	void setActiveFlag(bool value)
	{
		d_active = value;
	}
	bool getActiveFlag() const 
	{
		return d_active;
	}
private:
	AbstractSP() : d_object(0), d_size(0), d_active(false){}
protected:
	void* d_object;		//  
private:
	size_t d_size;		//  
	bool d_active;		//      
};

template <class T, class R, template <class> class S, template <class> class GC, template <class, template <class> class> class W> 
class SP :  public AbstractSP<R, S, GC, W>
					 
{
public:
	SP() : AbstractSP<R, S, GC, W>(new (W<AbstractSP<R, S, GC, W>, GC >::Allocate(sizeof(T))) T, sizeof(T)){}
	T* operator->()
	{
		return static_cast<T*>(d_object);
	}
	~SP()
	{
		static_cast<T*>(d_object)->~T();							//   
		W<AbstractSP<R, S, GC, W>, GC >::Deallocate(d_object);		//  
	}
private:
	SP(const SP&);
	SP& operator=(const SP&);
};

template <class T, class R, template <class> class S, template <class> class GC, template <class, template <class> class> class W>
class SH
{
	friend class WH<T, R, S, GC, W>;
public:
	SH() : d_flag(true), d_sp(new SP<T, R, S, GC, W>)
	{
		d_sp->AddRef();
	}
	SH(const SH& handle) : d_flag(true), d_sp(handle.d_sp)
	{
		d_sp->AddRef();
	}
	SH(const WH<T, R, S, GC, W>& handle) : d_flag(true)
	{
		if (!handle.d_sp)
			throw;
		d_sp = handle.d_sp;
		d_sp->AddRef();
	}
	template <class U> SH(const SH<U, R, S, GC, W>& handle) : d_flag(true)
	{
		{
			int unnamed[Conversion<U, T>::exists];
		}
		d_sp = const_cast<AbstractSP<R, S, GC, W>* >(handle.GetSP());
		d_sp->AddRef();
	}
	template <class U> SH(const WH<U, R, S, GC, W>& handle) : d_flag(true)
	{
		{
			int unnamed[Conversion<U, T>::exists];
		}
		if (!handle.GetSP())
			throw;
		d_sp = const_cast<AbstractSP<R, S, GC, W>* >(handle.GetSP());
		d_sp->AddRef();
	}
	SH& operator=(const SH& handle)
	{
		if (d_sp != handle.d_sp)
		{
			d_sp->Release();
			handle.d_sp->AddRef();
			d_sp = handle.d_sp;
			d_flag = handle.d_flag;
		}
		return *this;
	}
	template <class U> SH& operator=(const SH<U, R, S, GC, W>& handle)
	{
		{
			int unnamed[Conversion<U, T>::exists];
		}
		if (d_sp != handle.GetSP())
		{
			d_sp->Release();
			AbstractSP<R, S, GC, W>* tmp = const_cast<AbstractSP<R, S, GC, W>* >(handle.GetSP());
			tmp->AddRef();
			d_sp = tmp;
			d_flag = handle.GetFlag();
		}
		return *this;
	}

	template <class U> SH& operator=(const WH<U, R, S, GC, W>& handle)
	{
		{
			int unnamed[Conversion<U, T>::exists];
		}
		if (!handle.GetSP())
			throw;

		if (d_sp != handle.GetSP())
		{
			d_sp->Release();
			d_flag = false;
			d_sp = const_cast<AbstractSP<R, S, GC, W>* >(handle.GetSP());
		}
		return *this;
	}

	SP<T, R, S, GC, W>& operator->()
	{
		return *static_cast<SP<T, R, S, GC, W>*>(d_sp);
	}
	~SH()
	{
		if (d_flag)	d_sp->Release();
	}
	const AbstractSP<R, S, GC, W>* const GetSP() const
	{
		return d_sp;
	}
	bool GetFlag() const
	{
		return d_flag;
	}
private:
	AbstractSP<R, S, GC, W>* d_sp;
	bool d_flag;
};

template <class T, class R, template <class> class S, template <class> class GC, template <class, template <class> class> class W>
class WH
{
	friend class SH<T, R, S, GC, W>;
public:
	WH() : d_sp(0) {}
	WH(const SH<T, R, S, GC, W>& handle) : d_sp(handle.d_sp){}
	WH& operator=(const SH<T, R, S, GC, W>& handle)
	{
		d_sp = handle.d_sp;
		return *this;
	}
	template <class U> WH& operator=(const SH<U, R, S, GC, W>& handle)
	{
		{
			int unnamed[Conversion<U, T>::exists];
		}
		d_sp = const_cast<AbstractSP<R, S, GC, W>* >(handle.GetSP());
		return *this;
	}

	SP<T, R, S, GC, W>& operator->()
	{
		return *static_cast<SP<T, R, S, GC, W>*>(d_sp);
	}
	const AbstractSP<R, S, GC, W>* const GetSP() const
	{
		return d_sp;
	}
private:
	AbstractSP<R, S, GC, W>* d_sp;
};

template <class T>
class Iterator
{
public:
	virtual void operator++() = 0;
	virtual const T* operator()() = 0;
	virtual operator const void*() const = 0;
	virtual ~Iterator(){}
};

template <class T>
class GCObject
{
public:
	virtual Iterator<T>* iterator() const = 0;
	virtual ~GCObject(){}
};

#endif // _SP_H
